#  Copyright 2018 Ocean Protocol Foundation
#  SPDX-License-Identifier: Apache-2.0


cost = "service.main.cost"
service_license = "service.attributes.main.license"
sample = "service.attributes.additionalInformation.links.type"
categories = "service.attributes.additionalInformation.categories"
tags = "service.attributes.additionalInformation.tags"
created = "created"
dataToken = "dataToken"
datePublished = "service.attributes.main.datePublished"
dateCreated = "service.attributes.main.dateCreated"
updated_frequency = "service.attributes.additionalInformation.updateFrequency"
metadata_type = "service.attributes.main.type"
service_type = "service.type"

list_indexes = [cost, service_license, sample, categories, tags,
                created, dataToken, updated_frequency, service_type, metadata_type]



#  Copyright 2018 Ocean Protocol Foundation
#  SPDX-License-Identifier: Apache-2.0

from ssl import CERT_NONE, CERT_REQUIRED

from oceandb_driver_interface.utils import get_value
from pymongo import MongoClient, TEXT

from oceandb_mongodb_driver.indexes import list_indexes

_DB_INSTANCE = None


def get_database_instance(config_file=None):
    global _DB_INSTANCE
    if _DB_INSTANCE is None:
        _DB_INSTANCE = MongoInstance(config_file)

    return _DB_INSTANCE


class MongoInstance(object):

    def __init__(self, config=None):
        host = get_value('db.hostname', 'DB_HOSTNAME', 'localhost', config)
        port = int(get_value('db.port', 'DB_PORT', 27017, config))
        db_name = get_value('db.name', 'DB_NAME', 'db_name', config)
        collection = get_value('db.collection', 'DB_COLLECTION', 'collection_name', config)
        username = get_value('db.username', 'DB_USERNAME', None, config)
        password = get_value('db.password', 'DB_PASSWORD', None, config)

        ssl = get_value('db.ssl', 'DB_SSL', 'false', config)
        verify_certs = get_value('db.verify_certs', 'DB_VERIFY_CERTS', 'false', config)
        if verify_certs is False:
            ssl_cert_reqs = CERT_NONE
        else:
            ssl_cert_reqs = CERT_REQUIRED

        ca_cert_path = get_value('db.ca_cert_path', 'DB_CA_CERTS', None, config)
        client_key = get_value('db.client_key', 'DB_CLIENT_KEY', None, config)
        client_cert_path = get_value('db.client_cert_path', 'DB_CLIENT_CERT', None, config)

        if ssl == 'true':
            self._client = MongoClient(host=host,
                                       port=port,
                                       ssl=ssl,
                                       ssl_cert_reqs=ssl_cert_reqs,
                                       ssl_ca_certs=ca_cert_path,
                                       ssl_certfile=client_cert_path,
                                       ssl_keyfile=client_key)
        else:
            self._client = MongoClient(host=host,
                                       port=port
                                       )
        self._db = self._client[db_name]
        if username is not None and password is not None:
            print('username/password: %s, %s' % (username, password))
            self._db.authenticate(name=username, password=password)

        self._collection = self._db[collection]
        self._collection.create_index([("$**", TEXT)])
        for index in list_indexes:
            self._collection.create_index(index)

    @property
    def instance(self):
        return self._collection



#  Copyright 2018 Ocean Protocol Foundation
#  SPDX-License-Identifier: Apache-2.0

from datetime import datetime, timedelta
import pytest
from oceandb_driver_interface.oceandb import OceanDb
from oceandb_driver_interface.search_model import FullTextModel, QueryModel

from oceandb_mongodb_driver.utils import query_parser
from tests.ddo_example import ddo_sample

mongo = OceanDb('./tests/oceandb.ini').plugin


def delete_all():
    for r in mongo.list():
        mongo.delete(r['_id'])


def test_plugin_type_is_mongodb():
    assert mongo.type == 'MongoDB'


def test_plugin_write_and_read():
    did = 'did:ocn-asset:0x123456789abcdefghi#path1'
    mongo.write({'value': 'test'}, did)
    assert mongo.read(did)['_id'] == did
    assert mongo.read(did)['value'] == 'test'
    mongo.delete(did)


def test_update():
    mongo.delete(1)
    mongo.write({'value': 'test'}, 1)
    assert mongo.read(1)['value'] == 'test'
    mongo.update({'value': 'testUpdated'}, 1)
    assert mongo.read(1)['value'] == 'testUpdated'
    mongo.delete(1)


def test_plugin_list():
    delete_all()
    mongo.write({'value': 'test1'}, 1)
    mongo.write({'value': 'test2'}, 2)
    mongo.write({'value': 'test3'}, 3)
    assert mongo.list().count() == 3
    assert mongo.list()[0]['value'] == 'test1'
    mongo.delete(1)
    mongo.delete(2)
    mongo.delete(3)


def test_plugin_query():
    mongo.write(ddo_sample, ddo_sample['id'])
    search_model = QueryModel({'cost': ["0", "12"]})
    r = mongo.query(search_model)
    assert r[1] > 0, 'search query has no results.'
    assert r[0][0]['id'] == ddo_sample['id']

    search_model_2 = QueryModel({'cost': ["0", "12"], 'license': ['CC-BY']})
    assert mongo.query(search_model_2)[0][0]['id'] == ddo_sample['id']

    search_model_3 = QueryModel(
        {'cost': ["0", "12"], 'license': ['CC-BY'], 'type': ['access', 'metadata']})
    assert mongo.query(search_model_3)[0][0]['id'] == ddo_sample['id']

    search_model_4 = QueryModel({'sample': []})
    assert mongo.query(search_model_4)[0][0]['id'] == ddo_sample['id']

    search_model_5 = QueryModel({'created': ['today']})
    assert mongo.query(search_model_5)[0].retrieved == 0

    search_model_6 = QueryModel({'created': []})
    assert mongo.query(search_model_6)[0][0]['id'] == ddo_sample['id']

    search_model_7 = QueryModel({'text': ['weather']})
    assert mongo.query(search_model_7)[0][0]['id'] == ddo_sample['id']

    search_model_8 = QueryModel({'text': ['weather'], 'cost': ["0", "12"]})
    assert mongo.query(search_model_8)[0][0]['id'] == ddo_sample['id']

    search_model_9 = QueryModel({'dataToken': ['0x2eD6d94Ec5Af12C43B924572F9aFFe470DC83282']})
    assert mongo.query(search_model_9)[0][0]['id'] == ddo_sample['id']

    delete_all()


def test_plugin_query_text():
    total = 15
    mongo.write({'key': 'A', 'value': 'test first'}, 1)
    mongo.write({'key': 'B', 'value': 'test second'}, 2)
    mongo.write({'key': 'C', 'value': 'test third'}, 3)
    mongo.write({'key': 'D', 'value': 'test fourth'}, 4)
    search_model = FullTextModel('test', {'key': -1}, offset=3, page=1)
    search_model1 = FullTextModel('test', {'key': -1}, offset=3, page=2)
    assert mongo.text_query(search_model)[0].count(with_limit_and_skip=True) == 3
    assert mongo.text_query(FullTextModel('test'))[0].count(with_limit_and_skip=True) == 4
    assert mongo.text_query(search_model)[0][0]['key'] == 'D'
    assert mongo.text_query(search_model)[0][1]['key'] == 'C'
    assert mongo.text_query(search_model1)[0][0]['key'] == 'A'

    for i in range(5, total+1):
        value = 'test %s' % i
        mongo.write({'key': str(i), 'value': value}, i)

    offset = 4
    for page in range(1, 4):  # page: 1, 2, 3
        result = mongo.text_query(FullTextModel('test', offset=offset, page=page))
        assert result[1] == total
        assert result[0].count(with_limit_and_skip=True) == offset

    result = mongo.text_query(FullTextModel('test', offset=offset, page=4))
    assert result[0].count(with_limit_and_skip=True) == (total % offset)

    # Clean up
    for i in range(1, total):
        mongo.delete(i)


def test_query_parser():
    query = {'cost': [0, 10]}
    assert query_parser(query) == {"service.main.cost": {"$gte": 0, "$lte": 10}}
    query = {'cost': [15]}
    assert query_parser(query) == {"service.main.cost": {"$gte": 0, "$lte": 15}}
    query = {'type': ['access', 'metadata']}
    assert query_parser(query) == {
        "$and": [{"service.type": "access"}, {"service.type": "metadata"}]}
    query = {'license': []}
    assert query_parser(query) == {}
    query = {'license': [], 'type': ['access', 'metadata']}
    assert query_parser(query) == {
        "$and": [{"service.type": "access"}, {"service.type": "metadata"}]}
    query = {'license': ['CC-BY'], 'type': ['access', 'metadata']}
    assert query_parser(query) == {"$or": [{"service.attributes.main.license": "CC-BY"}], "$and": [
        {"service.type": "access"},
        {"service.type": "metadata"}]}
    query = {'created': ['today', 'lastWeek', 'lastMonth', 'lastYear']}
    assert query_parser(query)['created']['$gte'].year == (datetime.now() - timedelta(days=365)).year
    query = {'created': ['no_valid']}
    assert query_parser(query)['created']['$gte'].year == (
        datetime.now() - timedelta(weeks=1000)).year
    query = {'categories': ['weather', 'other']}
    assert query_parser(query) == {"$or": [{"service.attributes.additionalInformation.categories": "weather"},
                                           {"service.attributes.additionalInformation.categories": "other"}]}
    query = {'text': ['weather']}
    assert query_parser(query) == {"$text": {"$search": "weather"}}


def test_query_not_supported():
    query = {'not_supported': []}
    with pytest.raises(Exception):
        query_parser(query)


def test_default_sort():

    mongo.write(ddo_sample, ddo_sample['id'])
    ddo_sample2 = ddo_sample.copy()
    ddo_sample2['id'] = 'did:op:cb36cf78d87f4ce4a784f17c2a4a694f19f3fbf05b814ac6b0b7197163888864'
    ddo_sample2['service'][2]['attributes']['curation']['rating'] = 0.99
    mongo.write(ddo_sample2, ddo_sample2['id'])
    search_model = QueryModel({'cost': ["0", "12"]})
    assert mongo.query(search_model)[0][0]['id'] == ddo_sample2['id']
    mongo.delete(ddo_sample['id'])
    mongo.delete(ddo_sample2['id'])



"""Implementation of OceanDB plugin based in MongoDB"""
#  Copyright 2018 Ocean Protocol Foundation
#  SPDX-License-Identifier: Apache-2.0

import logging

from oceandb_driver_interface.plugin import AbstractPlugin
from oceandb_driver_interface.search_model import FullTextModel, QueryModel
from pymongo import DESCENDING

from oceandb_mongodb_driver.instance import get_database_instance
from oceandb_mongodb_driver.utils import query_parser

logger = logging.getLogger(__name__)


class Plugin(AbstractPlugin):
    """Mongo ledger plugin for `Ocean DB's Python reference
    implementation <https://github.com/oceanprotocol/oceandb-mongo-driver>`_.
    Plugs in a MongoDB instance as the persistence layer for Ocean Db
    related actions.
    """

    def __init__(self, config=None):
        """Initialize a :class:`~.Plugin` instance and connect to MongoDB.
        Args:
            *nodes (str): One or more URLs of MongoDB nodes to
                connect to as the persistence layer
        """
        self.driver = get_database_instance(config)
        self.logger = logging.getLogger('Plugin')
        logging.basicConfig(level=logging.INFO)

    @property
    def type(self):
        """str: the type of this plugin (``'MongoDB'``)"""
        return 'MongoDB'

    @staticmethod
    def _num_to_skip(page, num_per_page):
        return (page - 1) * num_per_page

    def write(self, obj, resource_id=None):
        if resource_id is not None:
            obj['_id'] = resource_id
        o = self.driver.instance.insert_one(obj)
        self.logger.debug('mongo::write::{}'.format(o.inserted_id))
        return o.inserted_id

    def read(self, resource_id):
        return self.driver.instance.find_one({"_id": resource_id})

    def update(self, obj, resource_id):
        prev = self.read(resource_id)
        self.logger.debug('mongo::update::{}'.format(resource_id))
        return self.driver.instance.replace_one(prev, obj)

    def delete(self, resource_id):
        self.logger.debug('mongo::delete::{}'.format(resource_id))
        return self.driver.instance.delete_one({"_id": resource_id})

    def list(self, search_from=None, search_to=None, limit=0):
        return self.driver.instance.find().limit(limit)

    def query(self, search_model: QueryModel):
        assert search_model.page >= 1, 'page value %s is invalid' % search_model.page

        query_result = self.driver.instance.find(query_parser(search_model.query))
        sort_params = [('service.attributes.curation.rating', DESCENDING)]
        if search_model.sort is not None:
            sort_params = list(search_model.sort.items())

        return (
            query_result.sort(sort_params)
            .skip(self._num_to_skip(search_model.page, search_model.offset))
            .limit(search_model.offset),
            query_result.count()
        )

    def text_query(self, full_text_model: FullTextModel):
        assert full_text_model.page >= 1, 'page value %s is invalid' % full_text_model.page

        find_params = {"score": {"$meta": "textScore"}}

        if full_text_model.sort is None:
            sort_params = [
                ('score', {'$meta': 'textScore'}),
                ('service.attributes.curation.rating', DESCENDING)]
        else:
            sort_params = list(full_text_model.sort.items())

        query_result = self.driver.instance.find(
            {"$text": {"$search": full_text_model.text}},
            find_params
        )

        return (
            query_result.sort(sort_params)
            .skip(self._num_to_skip(full_text_model.page, full_text_model.offset))
            .limit(full_text_model.offset),
            query_result.count()
        )



#  Copyright 2018 Ocean Protocol Foundation
#  SPDX-License-Identifier: Apache-2.0

from datetime import datetime

ddo_sample = {
    "@context": "https://w3id.org/future-method/v1",
    "id": "did:op:cb36cf78d87f4ce4a784f17c2a4a694f19f3fbf05b814ac6b0b7197163888865",
    "created": datetime.strptime("2016-02-08T16:02:20Z", '%Y-%m-%dT%H:%M:%SZ'),
    "dataToken": "0x2eD6d94Ec5Af12C43B924572F9aFFe470DC83282",
    "publicKey": [
        {
            "id": "did:op:b6e2eb5eff1a093ced9826315d5a4ef6c5b5c8bd3c49890ee284231d7e1d0aaa#keys-1",
            "type": "RsaVerificationKey2018",
            "owner": "did:op:6027c1e7cbae06a91fce0557ee53195284825f56a7100be0c53cbf4391aa26cc",
            "publicKeyPem": "-----BEGIN PUBLIC KEY...END PUBLIC KEY-----\r\n"
        },
        {
            "id": "did:op:d1fe2dc63e0e4fe2fff65d2077d71e39eef3ceda293a36265acc30c81d78ce95#keys-2",
            "type": "Ed25519VerificationKey2018",
            "owner": "did:op:4c27a254e607cdf91a1206480e7eb8c74856102316c1a462277d4f21c02373b6",
            "publicKeyBase58": "H3C2AVvLMv6gmMNam3uVAjZpfkcJCwDwnZn6z3wXmqPV"
        },
        {
            "id": "did:op:da1ec6bba5ab6842897599cb8e6c17b888e3fee1e8adb2ac18e30c6b511a78b8#keys-3",
            "type": "RsaPublicKeyExchangeKey2018",
            "owner": "did:op:5f6b885202ffb9643874be529302eb00d55e226959f1fbacaeda592c5b5c9484",
            "publicKeyPem": "-----BEGIN PUBLIC KEY...END PUBLIC KEY-----\r\n"
        }
    ],
    "authentication": [
        {
            "type": "RsaSignatureAuthentication2018",
            "publicKey": "did:op:0ebed8226ada17fde24b6bf2b95d27f8f05fcce09139ff5cec31f6d81a7cd2ea"
                         "#keys-1"
        },
        {
            "type": "ieee2410Authentication2018",
            "publicKey": "did:op:c24997ab64005abbe0bee5f6ad8dc4038c27d52b6e4a1f537a488441a9c2b0b9"
                         "#keys-2"
        }
    ],
    "proof": {
        "type": "UUIDSignature",
        "created": datetime.strptime("2016-02-08T16:02:20Z", '%Y-%m-%dT%H:%M:%SZ'),
        "creator": "did:example:8uQhQMGzWxR8vw5P3UWH1ja",
        "signatureValue": "QNB13Y7Q9...1tzjn4w=="
    },
    "service": [
        {
            "type": "access",
            "serviceEndpoint": "http://localhost:8030/api/v1/brizo/services/consume?pubKey"
                               "=0x00bd138abd70e2f00903268f3db08f2d25677c9e&agreementId"
                               "=0xeb4bb084942044a3857a5d107b48563a1ab56608c79342319697710336484fca&url=0",
            "index": "0",
            "main": {
                "cost": "10",
                "timeout": 0
            }
        },
        {
            "type": "compute",
            "serviceEndpoint": "http://mybrizo.org/api/v1/brizo/services/compute?pubKey=${"
                               "pubKey}&agreementId={agreementId}&algo={algo}&container={container}",
            "main": {
                "cost": "5",
                "timeout": 3600
            }
        },
        {
            "type": "metadata",
            "serviceEndpoint": "http://myaquarius.org/api/v1/provider/assets/metadata/{did}",
            "attributes": {
                "main": {
                    "name": "UK Weather information 2011",
                    "type": "dataset",
                    "dateCreated": datetime.strptime("2016-02-08T16:02:20Z", '%Y-%m-%dT%H:%M:%SZ'),
                    "author": "Met Office",
                    "license": "CC-BY",
                    "encoding": "UTF-8",
                    "compression": "zip",
                    "contentType": "text/csv",
                    "files": [
                        {
                            "url": "https://testocnfiles.blob.core.windows.net/testfiles/testzkp"
                                   ".pdf",
                            "checksum": "efb2c764274b745f5fc37f97c6b0e761",
                            "contentLength": "4535431",
                            "resourceId": "access-log2018-02-13-15-17-29-18386C502CAEA932"
                        }
                    ]

                },
                "curation": {
                    "rating": 0.93,
                    "numVotes": 123,
                    "schema": "Binary Votting"
                },
                "additionalInformation": {
                    "updateFrecuency": "yearly",
                    "structuredMarkup": [
                        {
                            "uri": "http://skos.um.es/unescothes/C01194/jsonld",
                            "mediaType": "application/ld+json"
                        },
                        {
                            "uri": "http://skos.um.es/unescothes/C01194/turtle",
                            "mediaType": "text/turtle"
                        }
                    ],
                    "links": [
                        {
                            "type": "sample",
                            "sample1":
                                "http://data.ceda.ac.uk/badc/ukcp09/data/gridded-land-obs/gridded"
                                "-land-obs-daily/"
                        },
                        {
                            "sample2":
                                "http://data.ceda.ac.uk/badc/ukcp09/data/gridded-land-obs/gridded"
                                "-land-obs-averages-25km/"
                        },
                        {
                            "fieldsDescription": "http://data.ceda.ac.uk/badc/ukcp09/"
                        }
                    ],
                    "inLanguage": "en",
                    "tags": ["weather", "uk", "2011", "temperature", "humidity"],
                    "description": "Weather information of UK including temperature and humidity",
                    "size": "3.1gb",
                    "copyrightHolder": "Met Office",
                    "categories": ["weather", "meteorology"],
                    "workExample": "423432fsd,51.509865,-0.118092,2011-01-01T10:55:11+00:00,7.2,68"
                }
            }
        }
    ]
}



#  Copyright 2018 Ocean Protocol Foundation
#  SPDX-License-Identifier: Apache-2.0

import logging
from datetime import datetime, timedelta

import oceandb_mongodb_driver.indexes as index

logger = logging.getLogger(__name__)

AND = "$and"
OR = "$or"
GT = "$gte"
LT = "$lte"


def query_parser(query):
    query_result = {}
    for key, value in query.items():
        if 'cost' == key:
            create_cost_query(query, query_result)
        elif 'license' == key:
            create_query(query['license'], index.service_license, query_result, OR)
        elif 'categories' == key:
            create_query(query['categories'], index.categories, query_result, OR)
        elif 'tags' == key:
            create_query(query['tags'], index.tags, query_result, OR)
        elif 'service_type' == key:
            create_query(query['service_type'], index.service_type, query_result, AND)
        # Support older key for searching service type
        elif 'type' == key:
            create_query(query['type'], index.service_type, query_result, AND)
        elif 'metadata_type' == key:
            create_query(query['metadata_type'], index.metadata_type, query_result, AND)
        elif 'updateFrequency' == key:
            create_query(query['updateFrequency'], index.updated_frequency, query_result, OR)
        elif 'created' == key:
            create_created_query(query, query_result)
        elif 'dataToken' == key:
            create_query(query['dataToken'], index.dataToken, query_result, AND)
        elif 'sample' == key:
            query_result[index.sample] = 'sample'
        elif 'text' == key:
            create_text_query(query['text'], query_result)
        else:
            logger.error('The key %s is not supported by OceanDB.' % key[0])
            raise Exception('The key %s is not supported by OceanDB.' % key[0])
    return query_result


def create_query(values, index_name, query, operator):
    for i, value in enumerate(values):
        if i == 0 and operator not in query:
            query[operator] = [{index_name: value}]
        else:
            query[operator] += [{index_name: value}]
    return query


def create_cost_query(query, query_result):
    if len(query['cost']) > 2:
        logger.info('You are sending more values than needed.')
    elif len(query['cost']) == 0:
        logger.info('You are not sending any value.')
    elif len(query['cost']) == 1:
        query_result[index.cost] = {GT: 0, LT: query['cost'][0]}
    else:
        query_result[index.cost] = {GT: query['cost'][0], LT: query['cost'][1]}
    return query_result


def create_created_query(query, query_result):
    now = datetime.now() - timedelta(weeks=1000)
    for values in query['created']:
        if values == 'today':
            now = datetime.now() - timedelta(days=1)
        elif values == 'lastWeek':
            now = datetime.now() - timedelta(days=7)
        elif values == 'lastMonth':
            now = datetime.now() - timedelta(days=30)
        elif values == 'lastYear':
            now = datetime.now() - timedelta(days=365)
        else:
            logger.info('The key %s is not supported in the created query' % values)
    query_result['created'] = {GT: now}
    return query_result


def create_text_query(query, query_result):
    query_result['$text'] = {"$search": query[0]}
    return query_result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

#  Copyright 2018 Ocean Protocol Foundation
#  SPDX-License-Identifier: Apache-2.0

from setuptools import setup, find_packages

with open('README.md', encoding='utf-8') as readme_file:
    readme = readme_file.read()

with open('CHANGELOG.md', encoding='utf-8') as changelog_file:
    changelog = changelog_file.read()

requirements = ['oceandb-driver-interface', 'pymongo', ]

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest', ]

setup(
    author="leucothia",
    author_email='devops@oceanprotocol.com',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    description="🐳 Mongo DB BigchainDB driver (Python).",
    install_requires=requirements,
    license="Apache Software License 2.0",
    long_description=readme,
    long_description_content_type='text/markdown',
    include_package_data=True,
    keywords='oceandb-mongodb-driver',
    name='oceandb-mongodb-driver',
    packages=find_packages(include=['oceandb_mongodb_driver']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/oceanprotocol/oceandb-mongodb-driver',
    version='0.2.1',
    zip_safe=False,
)
